#!/usr/bin/pwsh

# Settings

nv JAR_PATTERN           "*upgrade-tool-*.jar"
nv JAVA_BIN              "java"
nv REQUIRED_JAVA         "1.8.0"
nv DEFAULT_LOG_CFG_FILE  "upgrade-tool-log4j2.xml"

# Global variables

nv -Option "AllScope" JAVA_PATH                $null
nv -Option "AllScope" ACTION                   $null
nv -Option "AllScope" MULE_HOME_OUTDATED       $null
nv -Option "AllScope" MULE_HOME_UPDATED        $null
nv -Option "AllScope" OUTDATED_JAR             $null
nv -Option "AllScope" UPDATED_JAR              $null
nv -Option "AllScope" OLD_AGENT_VERSION        $null
nv -Option "AllScope" NEW_AGENT_VERSION        $null
nv -Option "AllScope" JAR_NAME                 $null
nv -Option "AllScope" DRY_RUN                  $false
nv -Option "AllScope" LOG_CFG_FILE             $null
nv -Option "AllScope" JAVA_ARGS                $null

# Strings

$Msg = @(0..20)
$Msg[0] = "ERROR: {0} version is {1}, minimum required is {2}, exiting"
$Msg[1] = "ERROR: Java binary not found neither in JAVA nor in JAVA_HOME, exiting"
$Msg[2] = "ERROR: Please use absolute paths for MULE_HOMEs, exiting" # <--- Not used in Windows
$Msg[3] = "INFO: Detected outdated version"
$Msg[4] = "INFO: Upgrading from {0} to {1}"
$Msg[5] = "WARN: Couldn't find any LOG4J configuration file, default configuration will be used"
$Msg[6] = "INFO: Skipping update, already at latest version ({0})"
$Msg[7] = "ERROR: MULE_HOME seems invalid, exiting"
$Msg[8] = "`r`nChecking Mule Agent"
$Msg[9] = "`r`nMule Runtime Upgrade Tool"
$Msg[10] = "INFO: Mule Agent is not present in the current setup, but it seems it was installed before, copying it to {0}"
$Msg[11] = "INFO: Mule Agent not installed, skipping setup"
$Msg[12] = "INFO: No update available, skipping update"
$Msg[13] = "ERROR: Unable to access jarfile"
$Msg[14] = "INFO: Running in dry-run mode, commands to be executed will be just displayed in stdout"

# Charset setup for Windows
chcp.com 1252 > $null

function New-Separator ( $Title ) {

  # Colorize
  [Console]::ForegroundColor = [System.ConsoleColor]::Green

  # Print
  $Char = "{0}" -f [char]0x2500
  $Underline = "{0}`r`n" -f ( $Char * $Title.Length )
  [Console]::Out.WriteLine( $Title )
  [Console]::Out.WriteLine( $Underline )

  # Reset color
  [Console]::ResetColor()
}

function Compare-Versions {

  # Set variables
  begin {
    $Err  = $false
    $NAME   = $Params[0]
    $CUR  = $Params[1].Split('.')
    $REQ  = $Params[2].Split('.')
  }

  # Compare versions
  process {
    if ( $CUR[0] -lt $REQ[0] ) {
      $Err = $true
    }
    elseif ( $CUR[0] -eq $REQ[0] ) {
      if ( $CUR[1] -lt $REQ[1] ) {
        $Err = $true
      }
      elseif ( $CUR[1] -eq $REQ[1] ) {
        if ( $CUR[2] -lt $REQ[2] ) {
          $Err = $true
        }
      }
    }
  }

  # Report error
  end {
    if ( $Err ) {
      return $true
    }
    else {
      return $false
    }
  }
}

function Search-in-Path {
  try {
    $p = Get-Command -Type "Application" $JAVA_BIN 2> $null
    if ( $? ) {
      $JAVA_PATH = $p[0].Path
      return $true
    }
    else {
      throw
    }
  }
  catch {
    return $false
  }
}

function Search-in-Env {
  try {
    if ( $env:JAVA_HOME ) {
      $JAVA_PATH = $env:JAVA_HOME + "/bin/" + $JAVA_BIN

      if ( Test-Path -Path "$JAVA_PATH" -PathType Leaf ) {
        return $true
      }

      else {
        throw
      }
    }
  }
  catch {
    return $false
  }
}

function Search-Java-Bin {
  if ( Search-in-Path ) {
    return $true
  }
  elseif ( Search-in-Env ) {
    return $true
  }
  else {
    [Console]::Error.WriteLine( $Msg[1] )
    exit 1
  }
}

function Check-JVM-Version {
  process {

    # Prepare the command to run

    $pinfo = New-Object System.Diagnostics.ProcessStartInfo
    $pinfo.FileName               = "$JAVA_PATH"
    $pinfo.Arguments              = "-version"
    $pinfo.UseShellExecute        = $false
    $pinfo.RedirectStandardError  = $true

    # Run the process and capture output

    $p = New-Object System.Diagnostics.Process
    $p.StartInfo = $pinfo
    $p.Start() > $null
    $p.WaitForExit()

    # Extract version number from the output

    $Version  = $( $p.StandardError.ReadToEnd() ).Split( [Environment]::NewLine )
    $Version  = [Regex]::Match( $Version[0], '(\d+).(\d+).(\d+)' ).Value
  }
  end {

    $Name     = "Java"
    $Cur      = "$Version"
    $Req      = "$REQUIRED_JAVA"
    $Params   = @( $Name, $Cur, $Req )

    $Err = Compare-Versions ( $Params )
    if ( $Err ) {
      $Msg = $Msg[0] -f $NAME, $Params[1], $Params[2]
      [Console]::Error.WriteLine( $Msg )
      exit 1
    }
  }
}

function Run-Validations {
  Search-Java-Bin > $null
  Check-JVM-Version
}

function Parse-Arguments {

  # Read parameters to extract the MULE_HOMEs

  for ( $i = 0; $i -lt $CmdArgs.Count; $i++ ) {
    switch -Regex ( $CmdArgs[$i] ) {

      "(rollback|upgrade|status)" {
        $ACTION = "$_"
        break
      }

      "^(-o|-o=*|--old-mule-distro|--old-mule-distro=*)" {
        if ( $_ -match "=" ) {
          $Param = $_
          $Param = $Param.Split("=")[1]
          $MULE_HOME_OUTDATED = $Param
        }
        else {
          $i++
          $MULE_HOME_OUTDATED = $CmdArgs[$i]
        }
        break
      }

      "^(-n|-n=*|--new-mule-distro|--new-mule-distro=*)" {
        if ( $_ -match "=" ) {
          $Param = $_
          $Param = $Param.Split("=")[1]
          $MULE_HOME_UPDATED = $Param
        }
        else {
          $i++
          $MULE_HOME_UPDATED = $CmdArgs[$i]
        }
        break
      }

      "^(-d|--dry-run)" {
        $DRY_RUN = $true
        break
      }

      "^(-l|--log-config=*)" {
        if ( $_ -match "=" ) {
          $Param = $_
          $Param = $Param.Split("=")[1]
          $LOG_CFG_FILE = $Param
        }
        else {
          $i++
          $LOG_CFG_FILE = $CmdArgs[$i]
        }
        break
      }
    }
  }
}

function Test-Homes {

  # Check that MULE_HOMEs are readable

  foreach ( $Path in "$MULE_HOME_OUTDATED", "$MULE_HOME_UPDATED" ) {

    # If it's not empty
    if ( $Path ) {

      # If it's not a directory
      if ( ! ( Test-Path -Path $Path -PathType Container ) ) {
        [Console]::Error.WriteLine( $Msg[7] )
        exit 1
      }
    }
  }
}

function Append-Java-Arg ( $Opts ) {
  # If JAVA_OPTS is not empty, append a whitespace
  $JAVA_ARGS += $( if ( $JAVA_ARGS ) {" "} ) + "$Opts"
  return $true
}

function Setup-Logging {
  if ( $LOG_CFG_FILE ) {
    $FILE = "$LOG_CFG_FILE"
  } else {
    $FILE = "$DEFAULT_LOG_CFG_FILE"
  }

  $Param = "$FILE" | Resolve-Path 2> $null
  if ( $Param ) {
    $Opts = "-Dlog4j2.configurationFile=" + $Param.Path
    Append-Java-Arg $Opts
  } else {
    [Console]::out.WriteLine( $Msg[5] )
    return $false
  }

  return $true
}

function Print-Dryrun-Msg ( $Line ) {
  [Console]::Out.WriteLine( "[ DRY-RUN ] {0}" -f $Line )
  return $true
}

function Restore-Agent {

  # After upgrading, mule-agent-plugin doesn't exist, so If that's empty, grab a copy from the backup

  $Src  = "$MULE_HOME_OUTDATED/.upgrade-tool-backup/*plugins/mule-agent-plugin" | Resolve-Path 2> $null
  $Dst  = "$MULE_HOME_OUTDATED/*plugins" | Resolve-Path 2> $null

  if ( $Src ) {
    $ValidSrc = Test-Path -Path "$Src"
  }

  if ( $Dst ) {
    $Obj = $Dst.Path + "/mule-agent-plugin"
    $ValidDst = Test-Path -Path "$Obj"
  }

  if ( $ValidSrc -and -not $ValidDst ) {
    $f = $Msg[10] -f $Dst
    [Console]::Out.WriteLine( $f )

    Copy-Item -Path $Src -Destination $Dst -Recurse -Force
  }
}

function Agent-Already-Exist {

  $Upgrade = $true

  $Obj = $ObjExist = $null
  $Obj = "$MULE_HOME_OUTDATED/*plugins/mule-agent-plugin/lib/mule-agent-api-*.*.*.jar" | Resolve-Path 2> $null
  if ( $Obj.Path ) {
    $ObjExist = Test-Path -Path $Obj.Path -PathType Leaf
    if ( ! $ObjExist ) {
      $Upgrade = $false
    }
    else {
      $OUTDATED_JAR = $Obj.Path
    }
  }
  else {
    [Console]::Out.WriteLine( $Msg[11] )
    $Upgrade = $false
  }

  $Obj = $ObjExist = $null
  $Obj = "$MULE_HOME_UPDATED/tools/agent-setup-*.*.*-amc-final.jar" | Resolve-Path 2> $null
  if ( $Obj.Path ) {
    $ObjExist = Test-Path -Path $Obj.Path -PathType Leaf
    if ( ! $ObjExist ) {
      $Upgrade = $false
    }
    else {
      $UPDATED_JAR = $Obj.Path
    }
  }
  else {
    [Console]::Out.WriteLine( $Msg[12] )
    $Upgrade = $false
  }

  return $Upgrade
}

function New-Agent-Version-Available {

  $OldVer = "$OUTDATED_JAR".Split("/")[-1]
  $OldVer = [Regex]::Match( $OldVer, '(\d+).(\d+).(\d+)' ).Value

  $NewVer = "$UPDATED_JAR".Split("/")[-1]
  $NewVer = [Regex]::Match( $NewVer, '(\d+).(\d+).(\d+)' ).Value

  $Name   = "Mule Agent"
  $Cur    = "$OldVer"
  $Req    = "$NewVer"
  $Params = @( $Name, $Cur, $Req )

  # Export info to the global variables:

  $OLD_AGENT_VERSION="$OldVer"
  $NEW_AGENT_VERSION="$NewVer"

  return Compare-Versions ( $Params )
}

function Upgrade-Agent {

  # These are the files we need to migrate manually
  $Files = @(
    "bin/amc_setup*"
    "tools/agent-setup*"
  )

  [Console]::Out.WriteLine( $Msg[3] )

  $f = $Msg[4] -f $OLD_AGENT_VERSION, $NEW_AGENT_VERSION
  [Console]::Out.WriteLine( $f )

  # Copy individual files manually

  if ( ! $DRY_RUN ) {
    Copy-Item -Recurse -Path $MULE_HOME_UPDATED/$Files[0] -Destination $MULE_HOME_OUTDATED/bin/
    Copy-Item -Recurse -Path $MULE_HOME_UPDATED/$Files[1] -Destination $MULE_HOME_OUTDATED/tools/
  } else {
    Print-Dryrun-Msg ( "Copy-Item -Recurse -Path {0} -Destination {1}" -f ( "$MULE_HOME_UPDATED/" + $Files[0] ), "$MULE_HOME_OUTDATED/bin/" )
    Print-Dryrun-Msg ( "Copy-Item -Recurse -Path {0} -Destination {1}" -f ( "$MULE_HOME_UPDATED/" + $Files[1] ), "$MULE_HOME_OUTDATED/tools/" )
  }

  # Run the upgrade, output is grayed out

  $CMD_ARGS = @( $Env:JAVA_OPTS, $JAVA_ARGS, "-jar", $UPDATED_JAR, "--update", "--mule-home", $MULE_HOME_OUTDATED)

  $pinfo = New-Object System.Diagnostics.ProcessStartInfo -Prop @{
    FileName                  = "$JAVA_PATH"
    Arguments                 = $CMD_ARGS
    UseShellExecute           = $false
    RedirectStandardOutput    = $false  # <--- Should be True, but that makes the script to hang, weird
    RedirectStandardError     = $false
  }

  if ( ! $DRY_RUN ) {
    [Console]::ForegroundColor = [System.ConsoleColor]::DarkGray
    $p = New-Object System.Diagnostics.Process
    $p.StartInfo = $pinfo
    $p.Start() > $null
    $p.WaitForExit()
    [Console]::ResetColor()
  } else {
    Print-Dryrun-Msg ( "{0} {1}" -f $JAVA_PATH, [String] $CMD_ARGS )
  }
}

function Setup-Mule-Agent {

  # If the Mule Agent is installed in the current Mule distro, check
  # if the new distro ships a newer version of the Mule Agent
  # If it does, update current Mule Agent by taking it from the newer
  # distro

  if ( $ACTION -eq "upgrade" ) {

    # When upgrading, we need both -o and -n

    if ( ! $MULE_HOME_OUTDATED -or ! $MULE_HOME_UPDATED ) {
      # No action taken
      return
    }

    New-Separator ( $Msg[8] )
    if ( $DRY_RUN ) {
      [Console]::Out.WriteLine( $Msg[14] )
    }
    Restore-Agent
    if ( Agent-Already-Exist ) {
      if ( New-Agent-Version-Available ) {
        Upgrade-Agent
      }
      else {
        $f = $Msg[6] -f $OLD_AGENT_VERSION
        [Console]::Error.WriteLine( $f )
      }
    }
    else {
      return
    }
  }
}

function Get-Jar-Name {

  $Obj = $JAR_PATTERN | Resolve-Path 2> $null
  if ( $Obj.Path ) {
    $ObjExist = Test-Path -Path $Obj.Path -PathType Leaf

    if ( ! $ObjExist ) {
      return $false
    } else {
      $JAR_NAME = $Obj.Path
      return $true
    }
  }
}

function Run-Program {

  New-Separator ( $Msg[9] )

  # If there are whitespaces, put them into an array
  if ( $CmdArgs -match "\s" ) {
    $CmdArgs = $CmdArgs.Split( "\s" )
  }

  if ( Get-Jar-Name ) {
    $CMD_ARGS = @( "$Env:JAVA_OPTS", "$JAVA_ARGS", "-jar", "$JAR_NAME" ) + $CmdArgs
    $pinfo = New-Object System.Diagnostics.ProcessStartInfo -Prop @{
      FileName            = "$JAVA_PATH"
      Arguments           = $CMD_ARGS
      UseShellExecute     = $false
      WorkingDirectory    = $PSScriptRoot
      }

    $p = New-Object System.Diagnostics.Process
    $p.StartInfo = $pinfo
    $p.Start() > $null
    $p.WaitForExit()
    if ( $p.ExitCode -eq 0 ) {
      return $true
    }
    return $false
  }
  else {
    [Console]::Error.WriteLine( $Msg[13] )
    exit $false
  }
}

function Main {
  Run-Validations
  Parse-Arguments
  Test-Homes
  Setup-Logging
  if ( Run-Program ) {
    Setup-Mule-Agent
  }
  else {
    return $false
  }
}

$CmdArgs = $args
if ( Main ) { exit 0 } else { exit 1 }
